@@ -1,18 +1,19 @@
 /*
-**  Copyright (c) 2007, 2008 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2007-2009 Sendmail, Inc. and its suppliers.
 **	All rights reserved.
 **  Sendmail, Inc. Confidential
 **
-**  $Id: batv-filter.c,v 1.32 2008/09/04 20:17:50 msk Exp $
+**  $Id: batv-filter.c,v 1.44 2009/03/27 19:38:11 msk Exp $
 */
 
 #ifndef lint
-static char batv_filter_c_id[] = "@(#)$Id: batv-filter.c,v 1.32 2008/09/04 20:17:50 msk Exp $";
+static char batv_filter_c_id[] = "@(#)$Id: batv-filter.c,v 1.44 2009/03/27 19:38:11 msk Exp $";
 #endif /* !lint */
 
 /* system includes */
 #include <sys/param.h>
 #include <sys/types.h>
+#include <sys/wait.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <netinet/in.h>
@@ -81,7 +82,7 @@ static char batv_filter_c_id[] = "@(#)$Id: batv-filter.c,v 1.32 2008/09/04 20:17
 # define SMFICTX	void
 # define _SOCK_ADDR	struct sockaddr
 
-int smfi_addheader __P((void *, char *, char *));
+int smfi_insheader __P((void *, int, char *, char *));
 int smfi_chgfrom __P((void *, char *, char *));
 void *smfi_getpriv __P((void *));
 char *smfi_getsymval __P((void *, char *));
@@ -89,7 +90,7 @@ int smfi_opensocket __P((int));
 void smfi_setconn __P((char *));
 void smfi_setdbg __P((int));
 void smfi_setpriv __P((void *, void *));
-void smfi_setreply __P((void *, char *, char *, char *));
+int smfi_setreply __P((void *, char *, char *, char *));
 
 char *smfis_ret[] = {
 	"SMFIS_CONTINUE",
@@ -99,8 +100,6 @@ char *smfis_ret[] = {
 	"SMFIS_TEMPFAIL",
 };
 
-typedef int bool;
-
 static void *fakepriv;				/* fake private space */
 #endif /* DEBUG */
 
@@ -171,11 +170,14 @@ typedef struct batv_domlist * DOMLIST;
 
 /* globals */
 bool dolog;
+bool die;
 bool addxhdr;
 bool notreally;
 bool bounceonly;
 bool skipauth;
 bool setreply;
+bool sendmail;
+int diesig;
 unsigned int version;
 char *progname;
 char *key;
@@ -190,6 +192,77 @@ DOMLIST domains;
 **  ============================= LOCAL STUFF =============================
 */
 
+/*
+**  BATV_SIGHANDLER -- signal handler
+**
+**  Parameters:
+**  	sig -- signal received
+**
+**  Return value:
+**  	None.
+*/
+
+static void
+batv_sighandler(int sig)
+{
+	if (sig == SIGINT || sig == SIGTERM || sig == SIGHUP)
+	{
+		diesig = sig;
+		die = TRUE;
+	}
+}
+
+/*
+**  BATV_STDIO -- set up the base descriptors to go nowhere
+**
+**  Parameters:
+**  	None.
+**
+**  Return value:
+**  	None.
+*/
+
+static void
+batv_stdio(void)
+{
+	int devnull;
+
+	/* this only fails silently, but that's OK */
+	devnull = open(_PATH_DEVNULL, O_RDWR, 0);
+	if (devnull != -1)
+	{
+		(void) dup2(devnull, 0);
+		(void) dup2(devnull, 1);
+		(void) dup2(devnull, 2);
+		if (devnull > 2)
+			(void) close(devnull);
+	}
+
+	(void) setsid();
+}
+
+/*
+**  BATV_KILLCHILD -- kill child process
+**
+**  Parameters:
+**  	pid -- process ID to signal
+**  	sig -- signal to use
+**  	dolog -- log it?
+**
+**  Return value:
+**  	None.
+*/
+
+static void
+batv_killchild(pid_t pid, int sig, bool dolog)
+{
+	if (kill(pid, sig) == -1 && dolog)
+	{
+		syslog(LOG_ERR, "kill(%d, %d): %s", pid, sig,
+		       strerror(errno));
+	}
+}
+
 /*
 **  BATV_INITCONTEXT -- initialize a context
 **
@@ -395,9 +468,9 @@ smfi_chgfrom(void *ctx, char *from, char *esmtp)
 }
 
 int
-smfi_addheader(void *ctx, char *hdr, char *val)
+smfi_insheader(void *ctx, int idx, char *hdr, char *val)
 {
-	printf("smfi_addheader(<ctx>, `%s', `%s')\n", hdr, val);
+	printf("smfi_insheader(<ctx>, %d, `%s', `%s')\n", idx, hdr, val);
 	return MI_SUCCESS;
 }
 
@@ -462,6 +535,14 @@ smfi_getsymval(void *ctx, char *sym)
 	snprintf(ret, l, "DEBUG-%s", sym);
 	return ret;
 }
+
+int
+smfi_setreply(void *ctx, char *smtp, char *esc, char *txt)
+{
+	printf("smfi_setreply(<ctx>, `%s', `%s', `%s')\n", smtp, esc, txt);
+
+	return MI_SUCCESS;
+}
 #endif /* DEBUG */
 
 /*
@@ -549,9 +630,10 @@ mlfi_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *addr)
 			if (bfc->ctx_internal)
 				break;
 		}
-		else if (hc->host_addr.s_addr != 0 &&
-		         hc->host_mask.s_addr != 0 &&
-		         bfc->ctx_addr.sa_family == AF_INET)
+
+		if (hc->host_addr.s_addr != 0 &&
+		    hc->host_mask.s_addr != 0 &&
+		    bfc->ctx_addr.sa_family == AF_INET)
 		{
 			struct sockaddr_in *sin;
 
@@ -768,7 +850,9 @@ mlfi_envfrom(SMFICTX *ctx, char **args)
 		char sender[BUFRSZ + 1];
 
 		/* continue if already signed */
-		if (strncmp(bfc->ctx_sender, "prvs=", 5) == 0)
+		if (!sendmail && strncmp(bfc->ctx_sender, "prvs=", 5) == 0)
+			return SMFIS_CONTINUE;
+		if (sendmail && strstr(bfc->ctx_sender, "+prvs=") != NULL)
 			return SMFIS_CONTINUE;
 
 		(void) time(&now);
@@ -781,9 +865,38 @@ mlfi_envfrom(SMFICTX *ctx, char **args)
 		SHA1_Update(&sha1, buf, strlen(buf));
 		SHA1_Final(digest, &sha1);
 
-		snprintf(sender, sizeof sender, "prvs=%1d%03d%02X%02X%02X=%s",
-		         version, expire, digest[0], digest[1], digest[2],
-		         bfc->ctx_sender);
+		if (sendmail)
+		{
+			char *at;
+			char tmp[BUFRSZ + 1];
+
+			sm_strlcpy(tmp, bfc->ctx_sender, sizeof tmp);
+
+			at = strchr(tmp, '@');
+			if (at != NULL)
+			{
+				*at = '\0';
+
+				snprintf(sender, sizeof sender,
+				         "%s+prvs=%1d%03d%02X%02X%02X@%s",
+				         tmp, version, expire, digest[0],
+				         digest[1], digest[2], at + 1);
+			}
+			else
+			{
+				snprintf(sender, sizeof sender,
+				         "%s+prvs=%1d%03d%02X%02X%02X",
+				         bfc->ctx_sender, version, expire,
+				         digest[0], digest[1], digest[2]);
+			}
+		}
+		else
+		{
+			snprintf(sender, sizeof sender,
+			         "prvs=%1d%03d%02X%02X%02X=%s",
+			         version, expire, digest[0], digest[1],
+			         digest[2], bfc->ctx_sender);
+		}
 
 		bfc->ctx_newsender = strdup(sender);
 		if (bfc->ctx_newsender == NULL)
@@ -814,16 +927,20 @@ mlfi_envfrom(SMFICTX *ctx, char **args)
 			sm_strlcat(esmtp, args[c], sizeof esmtp);
 		}
 
-		bfc->ctx_esmtp = strdup(esmtp);
-		if (bfc->ctx_esmtp == NULL)
+		if (esmtp[0] != '\0')
 		{
-			if (dolog)
+			bfc->ctx_esmtp = strdup(esmtp);
+			if (bfc->ctx_esmtp == NULL)
 			{
-				syslog(LOG_ERR, "%s: strdup(): %s",
-				       bfc->ctx_jobid, strerror(errno));
-			}
+				if (dolog)
+				{
+					syslog(LOG_ERR, "%s: strdup(): %s",
+					       bfc->ctx_jobid,
+					       strerror(errno));
+				}
 
-			return SMFIS_TEMPFAIL;
+				return SMFIS_TEMPFAIL;
+			}
 		}
 	}
 	
@@ -984,7 +1101,8 @@ mlfi_envrcpt(SMFICTX *ctx, char **args)
 		return SMFIS_CONTINUE;
 
 	/* if it's unsigned, complain */
-	if (strncmp(addr, "prvs=", 5) != 0)
+	if ((!sendmail && strncmp(addr, "prvs=", 5) != 0) ||
+	    (sendmail && strstr(addr, "+prvs=") == NULL))
 	{
 		/* only complain if it's a bounce, or we're being fascist */
 		if (strlen(bfc->ctx_sender) == 0 || !bounceonly)
@@ -1031,6 +1149,16 @@ mlfi_envrcpt(SMFICTX *ctx, char **args)
 		char buf[BUFRSZ + 1];
 		unsigned char digest[SHA1_DIGEST_SIZE];
 
+		if (sendmail)
+		{
+			char *plus;
+
+			plus = strstr(addr, "+prvs=");
+			assert(plus != NULL);
+
+			sig = plus + 6;
+		}
+
 		rcpt = (RCPTLIST) malloc(sizeof *rcpt);
 		if (rcpt == NULL)
 		{
@@ -1057,24 +1185,43 @@ mlfi_envrcpt(SMFICTX *ctx, char **args)
 			return SMFIS_TEMPFAIL;
 		}
 
-		eq = strchr(sig, '=');
-		if (eq == NULL)
+		if (sendmail)
 		{
-			if (dolog)
+			char *at;
+			char *plus;
+
+			at = strchr(sig, '@');
+			assert(at != NULL);
+
+			sm_strlcpy(orcpt, addr, sizeof orcpt);
+
+			plus = strstr(orcpt, "+prvs=");
+			assert(plus != NULL);
+			*plus = '\0';
+
+			sm_strlcat(orcpt, at, sizeof orcpt);
+		}
+		else
+		{
+			eq = strchr(sig, '=');
+			if (eq == NULL)
 			{
-				syslog(LOG_ERR,
-				       "%s: BATV syntax error in `%s'",
-				       bfc->ctx_jobid, addr);
+				if (dolog)
+				{
+					syslog(LOG_ERR,
+					       "%s: BATV syntax error in `%s'",
+					       bfc->ctx_jobid, addr);
+				}
+
+				free(rcpt->rcpt_orig);
+				free(rcpt);
+
+				return SMFIS_TEMPFAIL;
 			}
 
-			free(rcpt->rcpt_orig);
-			free(rcpt);
-
-			return SMFIS_TEMPFAIL;
+			sm_strlcpy(orcpt, eq + 1, sizeof orcpt);
 		}
 
-		sm_strlcpy(orcpt, eq + 1, sizeof orcpt);
-
 		if (!isdigit(sig[0]) ||
 		    !isdigit(sig[1]) ||
 		    !isdigit(sig[2]) ||
@@ -1131,10 +1278,29 @@ mlfi_envrcpt(SMFICTX *ctx, char **args)
 		SHA1_Update(&sha1, buf, strlen(buf));
 		SHA1_Final(digest, &sha1);
 
-		snprintf(tmp, sizeof tmp,
-		         "prvs=%1d%03d%02X%02X%02X=%s",
-		         sigversion, sigexpire, digest[0], digest[1],
-		         digest[2], orcpt);
+		if (sendmail)
+		{
+			char *at;
+			char tmpaddr[BUFRSZ + 1];
+
+			sm_strlcpy(tmpaddr, orcpt, sizeof tmpaddr);
+
+			at = strchr(tmpaddr, '@');
+			assert(at != NULL);
+			*at = '\0';
+
+			snprintf(tmp, sizeof tmp,
+			         "%s+prvs=%1d%03d%02X%02X%02X@%s",
+			         tmpaddr, sigversion, sigexpire,
+			         digest[0], digest[1], digest[2], at + 1);
+		}
+		else
+		{
+			snprintf(tmp, sizeof tmp,
+			         "prvs=%1d%03d%02X%02X%02X=%s",
+			         sigversion, sigexpire, digest[0], digest[1],
+			         digest[2], orcpt);
+		}
 
 		if (strcmp(tmp, addr) != 0)
 		{
@@ -1255,7 +1421,11 @@ mlfi_eom(SMFICTX *ctx)
 			       bfc->ctx_newsender);
 		}
 		else if (smfi_chgfrom(ctx, bfc->ctx_newsender,
+#if SMFI_VERSION >= 0x01000001
 		                      bfc->ctx_esmtp) != MI_SUCCESS)
+#else /* SMFI_VERSION >= 0x01000001 */
+		                      NULL) != MI_SUCCESS)
+#endif /* SMFI_VERSION >= 0x01000001 */
 		{
 			if (dolog)
 			{
@@ -1318,7 +1488,7 @@ mlfi_eom(SMFICTX *ctx)
 		         bfc->ctx_jobid != NULL ? bfc->ctx_jobid
 		                                : JOBIDUNKNOWN);
 
-		if (smfi_addheader(ctx, XHEADERNAME, xfhdr) != MI_SUCCESS)
+		if (smfi_insheader(ctx, 1, XHEADERNAME, xfhdr) != MI_SUCCESS)
 		{
 			if (dolog)
 			{
@@ -1490,7 +1660,7 @@ usage(void)
 {
 	fprintf(stderr, "%s: usage: %s -p socketfile -k key [options]\n"
 	                "\t-a allowaddrs  \tfile containing addresses to pass\n"
-	                "\t-A             \tskip mail from authenticated clients\n"
+	                "\t-A             \tenable auto-restart\n"
 	                "\t-d domainlist  \tdomain(s) to verify\n"
 	                "\t-D debuglevel  \tset milter debug level\n"
 	                "\t-f             \tdon't fork-and-exit\n"
@@ -1498,9 +1668,11 @@ usage(void)
 	                "\t-l             \tlog activity to system log\n"
 	                "\t-n             \tdon't actually do any filtering\n"
 	                "\t-P pidfile     \tfile to which to write pid\n"
+	                "\t-s             \tskip mail from authenticated clients\n"
 	                "\t-S             \trequest meaningful SMTP replies when rejecting\n"
 			"\t-u userid      \tchange to specified userid\n"
-	                "\t-V             \tprint version and terminate\n",
+	                "\t-V             \tprint version and terminate\n"
+	                "\t-x             \texperimental sendmail mode\n",
 	        progname, progname);
 	return EX_USAGE;
 }
@@ -1544,9 +1716,11 @@ struct smfiDesc smfilter =
 int
 main(int argc, char **argv)
 {
+	bool autorestart = FALSE;
 	bool dofork = TRUE;
 	int c;
 	int dbglevel;
+	int status;
 #ifndef DEBUG
 	int devnull;
 #endif /* ! DEBUG */
@@ -1567,8 +1741,10 @@ main(int argc, char **argv)
 
 	addxhdr = FALSE;
 	notreally = FALSE;
+	die = FALSE;
 	bounceonly = TRUE;
 	skipauth = FALSE;
+	sendmail = FALSE;
 	setreply = FALSE;
 	key = NULL;
 	allow = NULL;
@@ -1597,7 +1773,7 @@ main(int argc, char **argv)
 			break;
 
 		  case 'A':
-			skipauth = TRUE;
+			autorestart = TRUE;
 			break;
 
 		  case 'b':
@@ -1691,6 +1867,10 @@ main(int argc, char **argv)
 			pidfile = optarg;
 			break;
 
+		  case 's':
+			skipauth = TRUE;
+			break;
+
 		  case 'S':
 			setreply = TRUE;
 			break;
@@ -1711,6 +1891,10 @@ main(int argc, char **argv)
 			        BATV_PRODUCT, BATV_VERSION);
 			return 0;
 
+		  case 'x':
+			sendmail = TRUE;
+			break;
+
 		  default:
 			return usage();
 		}
@@ -2051,6 +2235,11 @@ main(int argc, char **argv)
 		values[n] = NULL;
 	}
 
+#ifdef DEBUG
+	return batv_debug();
+
+#else /* DEBUG */
+
 	/* Change user if appropriate */
 	if (user != NULL)
 	{
@@ -2085,15 +2274,181 @@ main(int argc, char **argv)
  
 	/* Activate logging */
 	if (dolog)
-#ifdef LOG_MAIL
+# ifdef LOG_MAIL
 		openlog(progname, LOG_PID, LOG_MAIL);
-#else /* LOG_MAIL */
+# else /* LOG_MAIL */
 		openlog(progname, LOG_PID);
-#endif /* LOG_MAIL */
+# endif /* LOG_MAIL */
 
-	(void) smfi_setconn(conn);
+	if (autorestart)
+	{
+		bool quitloop = FALSE;
+		int status;
+		pid_t pid;
+		pid_t wpid;
+		struct sigaction sa;
+
+		if (dofork)
+		{
+			pid = fork();
+			switch (pid)
+			{
+			  case -1:
+				if (dolog)
+				{
+					int saveerrno;
+
+					saveerrno = errno;
+
+					syslog(LOG_ERR, "fork(): %s",
+					       strerror(errno));
+
+					errno = saveerrno;
+				}
+
+				fprintf(stderr, "%s: fork(): %s\n",
+				        progname, strerror(errno));
+
+				return EX_OSERR;
+
+			  case 0:
+				batv_stdio();
+				break;
+
+			  default:
+				return EX_OK;
+			}
+		}
+
+		if (pidfile != NULL)
+		{
+			f = fopen(pidfile, "w");
+			if (f != NULL)
+			{
+				fprintf(f, "%ld\n", (long) getpid());
+				(void) fclose(f);
+			}
+			else
+			{
+				if (dolog)
+				{
+					syslog(LOG_ERR,
+					       "can't write pid to %s: %s",
+					       pidfile, strerror(errno));
+				}
+			}
+		}
+
+		sa.sa_handler = batv_sighandler;
+		sigemptyset(&sa.sa_mask);
+		sigaddset(&sa.sa_mask, SIGHUP);
+		sigaddset(&sa.sa_mask, SIGINT);
+		sigaddset(&sa.sa_mask, SIGTERM);
+		sa.sa_flags = 0;
+
+		if (sigaction(SIGHUP, &sa, NULL) != 0 ||
+		    sigaction(SIGINT, &sa, NULL) != 0 ||
+		    sigaction(SIGTERM, &sa, NULL) != 0)
+		{
+			if (dolog)
+			{
+				syslog(LOG_ERR, "[parent] sigaction(): %s",
+				       strerror(errno));
+			}
+		}
+
+		while (!quitloop)
+		{
+			if (batv_socket_cleanup(conn) != 0)
+			{
+				if (dolog)
+				{
+					syslog(LOG_ERR,
+					       "[parent] socket cleanup failed: %s",
+					       strerror(errno));
+				}
+				return EX_UNAVAILABLE;
+			}
+
+			pid = fork();
+			switch (pid)
+			{
+			  case -1:
+				if (dolog)
+				{
+					syslog(LOG_ERR, "fork(): %s",
+					       strerror(errno));
+				}
+
+				return EX_OSERR;
+
+			  case 0:
+				sa.sa_handler = SIG_DFL;
+
+				if (sigaction(SIGHUP, &sa, NULL) != 0 ||
+				    sigaction(SIGINT, &sa, NULL) != 0 ||
+				    sigaction(SIGTERM, &sa, NULL) != 0)
+				{
+					if (dolog)
+					{
+						syslog(LOG_ERR,
+						       "[child] sigaction(): %s",
+						       strerror(errno));
+					}
+				}
+
+				quitloop = TRUE;
+				break;
+
+			  default:
+				for (;;)
+				{
+					wpid = wait(&status);
+
+					if (wpid == -1 && errno == EINTR)
+					{
+						if (die)
+						{
+							batv_killchild(pid,
+							               diesig,
+							               dolog);
+
+							while (wpid != pid)
+								wpid = wait(&status);
+
+							if (pidfile != NULL)
+								(void) unlink(pidfile);
+
+							exit(EX_OK);
+						}
+					}
+
+					if (pid != wpid)
+						continue;
+
+					if (wpid != -1 && dolog)
+					{
+						if (WIFSIGNALED(status))
+						{
+							syslog(LOG_NOTICE,
+							       "terminated with signal %d, restarting",
+							       WTERMSIG(status));
+						}
+						else if (WIFEXITED(status))
+						{
+							syslog(LOG_NOTICE,
+							       "exited with status %d, restarting",
+							       WEXITSTATUS(status));
+						}
+					}
+
+					break;
+				}
+				break;
+			}
+		}
+	}
 
-#ifndef DEBUG
 	/* register with the milter interface */
 	if (smfi_register(smfilter) == MI_FAILURE)
 	{
@@ -2104,9 +2459,9 @@ main(int argc, char **argv)
 
 		return EX_UNAVAILABLE;
 	}
-#endif /* ! DEBUG */
 
 	/* try to establish the socket */
+	(void) smfi_setconn(conn);
 	if (smfi_opensocket(TRUE) == MI_FAILURE)
 	{
 		if (dolog)
@@ -2117,7 +2472,7 @@ main(int argc, char **argv)
 		return EX_UNAVAILABLE;
 	}
 
-	if (dofork)
+	if (!autorestart && dofork)
 	{
 		int save_errno;
 		pid_t pid;
@@ -2144,7 +2499,6 @@ main(int argc, char **argv)
 		}
 	}
 
-#ifndef DEBUG
 	/*
 	**  setsid() works on POSIX only.  Fortunately, that's
 	**  all we need for now.  Check sendmail/conf.c for a
@@ -2153,7 +2507,7 @@ main(int argc, char **argv)
 
 	(void) setsid();
 
-	if (!dofork)
+	if (!autorestart && !dofork)
 	{
 		int status;
 
@@ -2181,7 +2535,7 @@ main(int argc, char **argv)
 	}
 
 	/* write out the pid */
-	if (pidfile != NULL)
+	if (!autorestart && pidfile != NULL)
 	{
 		FILE *f;
 
@@ -2197,12 +2551,8 @@ main(int argc, char **argv)
 			       pidfile, strerror(errno));
 		}
 	}
-#endif /* ! DEBUG */
 
 	/* call the milter mainline */
-#ifdef DEBUG
-	return batv_debug();
-#else /* DEBUG */
 	if (dolog)
 	{
 		int status;
@@ -2236,6 +2586,12 @@ main(int argc, char **argv)
 		       BATV_VERSION, argstr);
 	}
 
-	return smfi_main();
+	status = smfi_main();
+
+	if (!autorestart && pidfile != NULL)
+		(void) unlink(pidfile);
+
+	return status;
+
 #endif /* DEBUG */
 }
