@@ -1,20 +1,20 @@
 /*
-**  Copyright (c) 2007, 2008 Sendmail, Inc. and its suppliers.
+**  Copyright (c) 2007-2009 Sendmail, Inc. and its suppliers.
 **	All rights reserved.
 **  Sendmail, Inc. Confidential
 **
-**  $Id: batv-filter.h,v 1.11 2008/09/05 18:22:52 msk Exp $
+**  $Id: batv-filter.h,v 1.16 2009/04/20 23:18:53 msk Exp $
 */
 
 #ifndef _BATV_FILTER_H_
 #define _BATV_FILTER_H_
 
 #ifndef lint
-static char batv_filter_h_id[] = "@(#)$Id: batv-filter.h,v 1.11 2008/09/05 18:22:52 msk Exp $";
+static char batv_filter_h_id[] = "@(#)$Id: batv-filter.h,v 1.16 2009/04/20 23:18:53 msk Exp $";
 #endif /* !lint */
 
 #define	BATV_PRODUCT	"Sendmail BATV Filter"
-#define	BATV_VERSION	"0.3.0"
+#define	BATV_VERSION	"0.4.0"
 
 #define	BATV_REJECTESC	"5.7.1"
 #define	BATV_REJECTSMTP	"550"
@@ -23,7 +23,7 @@ static char batv_filter_h_id[] = "@(#)$Id: batv-filter.h,v 1.11 2008/09/05 18:22
 #define	LOCALHOST	"localhost"
 #define	XHEADERNAME	"X-BATV"
 
-#define CMDLINEOPTS	"a:Abd:D:fhi:k:lnp:P:Su:V"
+#define CMDLINEOPTS	"a:Abd:D:fhi:k:lnp:P:sSu:Vx"
 
 #define	BUFRSZ		512
 #define	MAXARGV		2048
